@@ -12,6 +12,8 @@ use strict;
 use Graph::AdjacencyMap qw(:flags :fields);
 use base 'Graph::AdjacencyMap';
 
+use Scalar::Util qw(weaken);
+
 sub _new {
     my ($class, $flags, $arity) = @_;
     bless [ 0, $flags, $arity ], $class;
@@ -19,6 +21,13 @@ sub _new {
 
 require overload; # for de-overloading
 
+sub __strval {
+  my ($k, $f) = @_;
+  ref $k && ($f & _REF) &&
+    (($f & _STR) ? !overload::Method($k, '""') : overload::Method($k, '""')) ?
+	overload::StrVal($k) : $k;
+}
+
 sub __set_path {
     my $m = shift;
     my $f = $m->[ _f ];
@@ -32,7 +41,7 @@ sub __set_path {
     my @p = $p;
     my @k;
     my $k = shift;
-    my $q = ref $k && ($f & _REF) && overload::Method($k, '""') ? overload::StrVal($k) : $k;
+    my $q = __strval($k, $f);
     push @k, $q;
     return (\@p, \@k);
 }
@@ -73,7 +82,7 @@ sub __has_path {
     my @p = $p;
     my @k;
     my $k = shift;
-    my $q = ref $k && ($f & _REF) && overload::Method($k, '""') ? overload::StrVal($k) : $k;
+    my $q = __strval($k, $f);
     push @k, $q;
     return (\@p, \@k);
 }
